/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "common.h"
#include "parse.h"
#include "atsend.h"

#define TMPDISCONN_TIME 60
#define ALARM_CLEAN 	0
#define ALARM_CHECK 	1

static void mainRoutine     (void);
static int  execCmd 	    (int type, char *command);
static int  handleCmdByKey  (type_key* k);
static int  isBemusedCommand(char *cmd);

void        freeTimers      ();
int         verifyTimers    (int ticksInSec);
int         cancelTimer     (char *id);
timerCmd*   findTimer       (char *id, timerCmd **prev);
int         processTimerCmd (char *origCmd);
int         createTimer     (char* id, int type, int tmout, int times, char *cmd);

char tmp	    [MAXMAXLEN];
char logfile	    [MAXLEN];
char callerId	    [MAXLEN];
char modifierString [MAXARGLEN];
char tmptext        [MAXCMDLEN];


int  alarmTimer   = 0;
int  repeatTimer  = 0;
int  callTimer	  = 0;
int  commandTimer = 0;
int  paramValue   = -1;

int  useCallId    = 1;
int  remoteOn     = 1;
int  repeatNow    = 0;
type_key*  repeatCmdPtr = NULL;
int  callOn       = 0;
int  model        = MODEL_DEFAULT;

timerCmd* timers  = NULL;

int  serverMode   = SERVER_NO;  // By default use AT commands and devices in /dev
                                // If set to 1 or 2, then do not use AT command and use socket 
int  port         = 0;
int  initialized  = 0;

void logger(char *head, char *str)
{
    FILE *fplog;
    time_t logtime;
    char *timestr;
 
    if (getLog()) {
    	
	if (strcmp(head, "DEBUG") == 0 && getDebug() == 0) {
	    return;
	}
    
    	fplog = fopen(logfile, "a");
    	if (fplog!=NULL) {
            if (strcmp(head, "CFG") != 0) {
   	        time(&logtime);
   	        timestr = (char *) ctime(&logtime);
   	        timestr[strlen(timestr)-1]=0;
   	        fprintf(fplog, "[%s] - ",timestr);
	    }	
	    if(head!=NULL) {
		fprintf(fplog, "%s - ",head);
	    }
	    fprintf(fplog, "%s\n",str);
	    
	    fclose(fplog);
    	} else {
	    printf("Can't open log file >%s<\n", logfile);
    	}
    }
}

int openPort(char * portIn) 
{
    logger("INFO", "openPort()");
    char* strPort;
    int ret = 0;
    
    // device should be in format:
    // 1. /dev/something
    // 2. socket:5000
    // 3. bluetooth:10 or bluetooth (port will be 10 by default)

    if ((strPort = strstr(portIn, INET_SOCKET)) != NULL) {
        
        strPort += strlen(INET_SOCKET);

        port = atoi(strPort);

        sprintf(tmp, "TCP Server mode. Use port %d\n", port);
        logger("DEBUG", tmp);
    
        if ((ret = openSocketPort(SERVER_TCP, port)) < 0) {
            return -1;
        }
    
        serverMode = SERVER_TCP;

    } else if ((strPort = strstr(portIn, BT_SOCKET)) != NULL) {
 
        strPort += strlen(BT_SOCKET);
        
        if (strstr(strPort, ":") == NULL) { // just "bluetooth
            port = 10;
        } else {
            strPort++;
            port = atoi(strPort);
        }
        sprintf(tmp, "Bluetooth Sockets Server mode. Use channel %d", port);
        logger("DEBUG", tmp);
    
        if ((ret = openSocketPort(SERVER_BT, port)) < 0) {
            return -1;
        }
    
        serverMode = SERVER_BT;

    } else {
    
        sprintf(tmp, "Serial Client mode. Use device %s", portIn);
        logger("DEBUG", tmp);

        if (bt_open_port(portIn) < 0) {
            printf("ERROR: open port\n");
            return -1;
        }
    }
    
    if (initialized == 0) { 
        // Run "onInit" command
        int i = 0;
        type_key *k = findItem(EVT_INIT, &i);
        if (k && i == 1) {
            logger("INFO", "Exec cmd on init");
            handleCmdByKey(k);  // Since we not connected yet we did not interested in return code
        } else {
            logger("INFO", "No command on init\n");
        }
        initialized++;
    }
    return 1;
}

// Then port will be closed from a forked child use 0
static void closePort(int final) 
{
    if (serverMode > 0) {
        closeSocketPort();
    } else {
        if (bt_close_port(final) < 0) {
            printf("ERROR: close port\n");
        }
    }
}

int main(int argc, char *argv[])
{
    int i;
    for (i=1; i<argc; i++)
    {
    	if (strcmp(argv[i],"-f")==0 && i+1<argc) {
	    break;
	}
    }
	
    if (init_cfg(i<argc ? argv[++i] : NULL)!=1) {
    	printf("ERROR: Fails in init_cfg()\n");
	freeCfg();
    	exit(1);
    }
		
    if (parse_opts(argc, argv)!=1 ) {
    	printf("ERROR: Fails in parse_opts()\n");
	freeCfg();
    	exit(1);
    }
    if (getLog()) {
    	strcpy(logfile, LOGFILE);
	
    	printf("INFO: log file is %s\n",logfile);
		
    	// Just  truncate file
    	FILE *fplog = fopen(logfile, "w");
    	fclose(fplog);
		
    	printConf();
    }
    
    // Not to core dump if connection will close
    signal(SIGPIPE, SIG_IGN);   
    signal(SIGCHLD, SIG_IGN);   
    
    // main cycle
    mainRoutine();

    return 0;
}

// This is not clean-up function. It just finished the connection.
void finishAll()
{
    char *ptr;
    char cmd[32];
    char answer[1024];
	
    // finish events forwarding
    printf("Finish events forwarding\n");
    logger("INFO","Finish events forwarding");
    
    if (serverMode == 0) {
        ptr = getAT_CMER(CMER_OFF);
        if (ptr == NULL) {
    	    if (model == MODEL_MOTOROLA) {
	        strcpy(cmd,DEF_AT_CMER_OFF_MOTOROLA);
    	    } else if (model == MODEL_SE) {
	        strcpy(cmd,DEF_AT_CMER_OFF_SE);
    	    } else if (model == MODEL_SAGEM) {
	        strcpy(cmd,DEF_AT_CMER_OFF_SAGEM);
	    } else {
	        strcpy(cmd,DEF_AT_CMER_OFF_DEFAULT);
	    }
        } else {
    	    strcpy(cmd,ptr);
        }
        strcat(cmd,"\r");
        sendSyncCommand(cmd, answer, sizeof(answer), NULL);
    
        //handle_command(answer); // Does it needed ?
    
        // and resend No back
        //sendSyncCommand("AT+CKPD=\"e\"\r", answer, sizeof(answer), NULL);    // Does it needed ?
    }
    closePort(1);
    
    return;
}

static char *readResultsFile() 
{
    FILE *fp;

    fp=fopen(TOFILE,"r");
    if (fp == NULL) return NULL;
    
    int n = fread(tmptext, sizeof(char), MAXCMDLEN-1, fp);
    tmptext[n] = '\0';
    
    return tmptext;
}

static int processCommands(cmdItem* cmdList) 
{
    while (cmdList) {
            
        int ret = execCmd(cmdList->type, cmdList->descr);
        if (ret != EXIT_OK) {
            sprintf(tmp, "Return code from execCmd() is %d",ret);
            logger("DEBUG", tmp);
 	     
            if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
  	        logger("DEBUG", "Clean auto repeat flag after execCmd()");
 	        repeatNow    = 0;
 	        repeatCmdPtr = NULL;
		
	        return ret;
            }
        }
        cmdList = cmdList->next;
    }
    return EXIT_OK;
}

int handleCmdByKey(type_key* k)
{
    if (k == NULL) {
    	logger("DEBUG", "handleCmdByKey() NULL input");
	return EXIT_OK;
    }

    sprintf(tmp, "handleCmdByKey() >%s<",k->key);
    logger("DEBUG", tmp);
    
    cmdItem* cmdList = getCommand(k);
    if (cmdList == NULL) {
    	logger("DEBUG", "No appropriate command was found by getCommand()");
    } else {
     	sprintf(tmp, "ready to execute command");
     	logger("DEBUG", tmp);
        
        // In AT mode: User had pressed some key ... 
        // Send ToMainMenu sequence of CKPD's to show main menu again
        if (serverMode == SERVER_NO) {
            if (sendSeq(getToMainMenu()) != EXIT_OK) {
    	        logger("DEBUG","Fails in sendSeq()");
            }
        }
        
        return processCommands(cmdList);
    }	
	
    return EXIT_OK;	
}

int handle_key_press(char* keyIn) 
{
    int ret;
    
    sprintf(tmp, "handle_key_press() >%s<", keyIn);
    logger("DEBUG", tmp);

    // Search an alias
    char *key = NULL;
    if((key = findAlias(keyIn)) == NULL) {
        key = keyIn;
    }    
    
    // If modifier was set already it needs to verify which sequence we go
    if (modifierString[0] != '\0') {
    	strcat(modifierString," "); // Add one space first
    	strcat(modifierString, key);
    	sprintf(tmp, "Modifier+sequence is: >%s<", modifierString);
    	logger("DEBUG", tmp);
    }
    
    // Execute appropriate command
    int i;
    type_key *k = findItem((modifierString[0] == '\0' ? key : modifierString), &i);
    if (k && i == 1) {
    	if ((!repeatNow) && getAutoRepeat()) {
 	    logger("DEBUG", "Set auto repeat flag");
 	    repeatNow    = 1;
 	    repeatCmdPtr = k;
     	} 
	
	ret = handleCmdByKey(k);
	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    return ret;
    	}
	    
    	// Clean modifier string
    	modifierString[0] = '\0';
	    
    } else if (k && i == 0){
    	logger("DEBUG", "Got part of multi keys sequence. Nothing to do.");
    	if (modifierString[0] == '\0') {
	    logger("DEBUG", "Start of multi key sequence");
    	    strcpy(modifierString, key);
    	} 

        if(serverMode == SERVER_NO) {
            ret = sendSeq(getToMainMenu());
	    if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	        return ret;
    	    }
        }
	    
    } else {
    	// User had pressed some key ... 
    	// Send ToMainMenu sequence of CKPD's to show main menu again
    	if (serverMode == SERVER_NO && sendSeq(getToMainMenu()) != EXIT_OK) {
    	    logger("DEBUG","Fails in sendSeq()");
    	}
	
    	logger("DEBUG", "No approprite key definition was found by findItem()");
    	if (modifierString[0] != '\0') {
	    // Clean modifier string
    	    modifierString[0] = '\0';
	    logger("DEBUG", "Clean modifier string");
    	}
    }
    
    return EXIT_OK;
}

int handle_command(char* cmd) 
{
    char *prev, *next, *key;
    int ret;

    paramValue = -1;

    //skip lines starting with \n and \r
    if (cmd[0] == '\r') cmd++;
    if (cmd[0] == '\n') cmd++;

    sprintf(tmp, "handle_command() >%s<", cmd);
    logger("DEBUG", tmp);
  
    // most common case
    if (!cmd[0]) {
    	//logger("DEBUG", "handle_command() skip 1");
	return EXIT_OK;
    }
    
    // if recieved multiline command - handle line by line and return
    prev = cmd;
    next = strchr(cmd, '\r');
    if (next == NULL && serverMode > 0) {   // Java client will send +CKEV: 1,1; +CKEV: 1,0
    	next = strchr(cmd, '\r');
    }
    if (next) {
    	logger("DEBUG", "handle_command() multiline");
	
        char copy[1024];
        int len;
    
        do {     
            len = next-prev;
            
            if (len > 2) {
                memcpy(copy, prev, len);
                copy[len] = 0;
                
                // use recursion
                ret = handle_command(copy);
	    	if (ret == EXIT_DISCON || ret == EXIT_ABORT) {
	    	    return ret;
	    	}
            }
            prev = next+1;
        } while ((next = strchr(prev, '\r'))) ;
    
        // now return
        return EXIT_OK;
    }
    
    if (IS_OK(cmd)) {	// OK - nothing to do
        return EXIT_OK;
    }
    
    // Handle different keypresses
    // Motorola: +CKEV: "1",1
    // SE      : +CKEV: 1,1
    // Sagem: in rare ) cases it could be (between > and <) >,1:< ?
    
    if (memcmp(cmd, DEF_CKEV, 6) == 0) {
       		
	int isSE = MODEL_SE;
	
        key = cmd+7;
	
	// Sagem test
	// key+=3;
	
	if (*key == '"') {	// Seems this is Motorola
	    key++;
	    //logger("DEBUG","+CKEV is in Motorola format");
	    isSE = MODEL_MOTOROLA;
	}

	if (*key == ',') {	// Seems this is Sagem
	    logger("DEBUG","+CKEV is in rare Sagem format (Empty code)");
	    isSE = MODEL_SAGEM;
	}

	int i = 1;
	
	if (isSE == MODEL_SE) {	    	// SonyEricsson & default
	    while(key[i] != ',') i++;
	    key[i] ='\0';
	    
	    i = i+1;
	} else if (isSE == MODEL_MOTOROLA) {	// Motorola
	    while(key[i] != '"') i++;
	    key[i] ='\0';

	    i = i+2;
	} else if (isSE == MODEL_SAGEM) {	// rare Sagem case
	    *key ='\0';
     	} else {
	    logger("ERROR","Can't recognize +CKEV event !!!");
	    return EXIT_NOK;
	}
	
	if (key[i] == '1') {
	    logger("DEBUG","Button down event");
            return handle_key_press(key);
    	} else if (key[i] == '0') {
	    logger("DEBUG","Button up event");
            // In general we skip this
	    if (repeatNow) {
	    	// Clear auto repeat flag
		logger("DEBUG","Clean auto repeat flag");
    	    	repeatNow    = 0;
    	    	repeatCmdPtr = NULL;
	    }
	    return EXIT_OK;
    	} else {
    	    sprintf(tmp, "something wrong: key should be pressed or released: cmd=>%s< char=>%c<",cmd, key[i]);
            logger("DEBUG",tmp);
            return EXIT_NOK;
    	}
    } else if (memcmp(cmd, DEF_CLCC,6) == 0) {
	logger("INFO","Caller ID received. Skip it.");

	return EXIT_OK;
	
    } else if (strstr(cmd, DEF_RING) != NULL) {     // Incoming call
	logger("INFO","Incoming call");     	    // This event sent periodically until used answered a call
	
	if (callOn == 0) {
	    if (useCallId) {
	    	logger("DEBUG", "Get CallerId on incoming call");
    	    	ret = getClip(callerId);

		sprintf(tmp, "CallerId is %s",callerId);
	    	logger("INFO", tmp);
		
	    	if(ret == EXIT_EXACT) { // Active call exists
		    logger("DEBUG","Caller ID received");
	    	} else { 
		    logger("DEBUG","Can't get Caller ID. Probably call was finished");
		}

	    }
	    //
	    // May be we can get +CLIP: "+xxxxxxxxxxxx",145,,,,0< here ?
	    // 

    	    int i = 0;
	    type_key* k = findItem(EVT_INCALL, &i);
	
    	    if (k && i == 1) {
	    	logger("DEBUG", "Exec cmd on incoming call");
  	    	ret = handleCmdByKey(k);
	    	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    return ret;
    	    	}
	    }

	    // If there is (EndCall) command,then set timer to determine when call ends
            i = 0;
	    k = findItem(EVT_ENDCALL, &i);
	    if (k && i == 1) {
	    	logger("DEBUG", "Set call timer");
	    	callTimer = 0;
	    	callOn    = 1;
	    }
	    
	    return EXIT_OK;
    	} else {
	    logger("DEBUG","Skip it."); // This event sent periodically until used answered a call
    	}
    } else if (isBemusedCommand(cmd)) {     	// Bemused server emulation tricks
    	return EXIT_OK;
    }
    
    sprintf(tmp, "Unhandled cmd %s", cmd);
    logger("DEBUG",tmp);
    
    return EXIT_OK;
}

static int read_command(char* buf, int l) 
{
    int ch, len = 0;
    char* p = buf;
    if (serverMode) {
        len = readSocketConn(buf, l);
        if (len == EOF) {
            buf[0] = 0;
            return EOF;
        }
    } else {
        while (len < l) {
            ch = bt_readchar(100);
            if (ch >= 0 && ch != '\r') {
                *p++ = (char)ch;
                len++;
            } else  if (ch == EOF) {
    	        buf[0] = 0;
                return EOF;
            } else {
                break;
            }
        }
    }
  
    buf[len] = 0;
    return len;
}

// mode: 0 - clear alarms
//       1 - if fired - process, then clear alarms
void manageAlarms(int mode)
{
    char    *pf;
    cmdItem *pg;
    int i;
    struct stat buf;
        
    int n = getNumAlarms();
    for(i=0;i<n;i++) {
    	pf = getAlarmInfo(i, &pg);
	
    	// If file exists
	int fs = stat(pf, &buf);
	if (fs == 0) {
	    if (mode == ALARM_CLEAN) {
	    	logger("DEBUG","Remove file");
	    } else {    	
    	    	sprintf(tmp, "Alarm %s fired", pf);
    	    	logger("INFO",tmp);
		
                processCommands(pg);
 	    }
            remove(pf);
	}
    }
    return;
}

int initPort()
{
    char answer[1024];
    char cmd[32];
 
    logger("DEBUG","initPort");
 
    if (serverMode) {
        // Wait for incoming connection
        logger("INFO", "Server mode: Waiting connection");
        return listenAndAcceptSocketConn();
    } 
 
    answer[0] = '\0';
    sendSyncCommand("ATZ\r", answer, sizeof(answer), NULL);
    //sendSyncCommand("AT\r", answer, sizeof(answer), NULL);
    if(!IS_OK(answer)) {
    	printf("ERROR: AT -> %s\n", answer);
    	return -1;
    }
 	
    sendSyncCommand("ATE0\r", answer, sizeof(answer), NULL);
    if(!IS_OK(answer)) {
    	printf("ERROR: ATE0 -> %s\n", answer);
    	return -1;
    }
    char* ptr = getCharset();
    if (ptr) {
    	sprintf(cmd,"%s\"%s\"\r", DEF_AT_CSCS, ptr); 
    	sendSyncCommand(cmd, answer, sizeof(answer), NULL);
    	if(!IS_OK(answer)) {
    	    printf("ERROR: Can't set charset to %s\n", ptr);
    	}     
    }

    sprintf(cmd,"%s\r", DEF_AT_CLIP); 
    sendSyncCommand(cmd, answer, sizeof(answer), NULL);
    if(!IS_OK(answer)) {
    	printf("ERROR: Can't set CLIP ON\n");
    }
	
    
    ptr = getAT_CMER(CMER_ON);
    if (ptr == NULL) {	    // We have to determine model
    	sprintf(cmd,"%s\r", DEF_AT_CGMI); 
    	sendSyncCommand(cmd, answer, sizeof(answer), NULL);
	
	if (strstr(answer,STR_MOTOROLA)) {
	    model = MODEL_MOTOROLA;
	    strcpy(cmd,DEF_AT_CMER_ON_MOTOROLA);
	} else if (strstr(answer,STR_SE)) {
	    model = MODEL_SE;
	    strcpy(cmd,DEF_AT_CMER_ON_SE);
	} else if (strstr(answer,STR_SAGEM)) {
	    model = MODEL_SAGEM;
	    strcpy(cmd,DEF_AT_CMER_ON_SAGEM);
	} else if (strstr(answer,STR_SIEMENS)) {
	    model = MODEL_SIEMENS;
	    strcpy(cmd,DEF_AT_CMER_ON_SIEMENS);
	} else {
	    model = MODEL_DEFAULT;
	    strcpy(cmd,DEF_AT_CMER_ON_DEFAULT);
	}
    
    } else {
    	strcpy(cmd,ptr);
    }
    strcat(cmd,"\r");
    sendSyncCommand(cmd, answer, sizeof(answer), NULL);
    if(!IS_OK(answer)) {
    	printf("ERROR: AT_CMER SET >%s< -> %s\n", cmd, answer);
    	return -1;
    }

    // Siemens S55 needs additional AT+CMEC=2 to make AT+CKPD works
    // not sure about other Siemens phones
    if (model == MODEL_SIEMENS) {
    	sprintf(cmd,"%s\r", DEF_AT_CMEC); 
	sendSyncCommand(cmd, answer, sizeof(answer), NULL);
    	if(!IS_OK(answer)) {
    	    printf("ERROR: ON AT+CMEC\n");
	}
    }
    
    // Will set global callerId	[MAXLEN];
    int ret = getClip(callerId);
    if(ret == EXIT_NOK) {	// Got ERROR on AT+CLCC; probably phone does not supports this command
    	printf("ERROR: fails in getClip\n");
	useCallId = 0;
    }
    return 1;
}

static void mainRoutine(void) 
{
    char answer[1024];
    int ret = EXIT_OK;
    int retRead, i;
    type_key *k = NULL;
   
    logger("DEBUG","mainRoutine");
    
    int dmn = autoConnect();
    int rs  = getRetrySecs();
    strcpy(callerId, "NO CALLER ID");
    
    while (1){
        
    	// try to open port
        if(openPort(getDevice()) == 1 &&   // Open device
           initPort()            == 1 ) {  // Init modem: AT, ATE0, AT+CMER, in server mode waits for connection
            
            manageAlarms(ALARM_CLEAN);
	    
            i = 0;    
	    k = findItem(EVT_CONNECT, &i);
    	    if (k && i == 1) {
    	        logger("INFO", "Exec cmd on connect");
    	        ret = handleCmdByKey(k);
    	        if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    return;
    	        }
    	    } else {
    	        printf("Start event forwarding\n");
    	    }

            logger("INFO","Start event forwarding");
            
            while (1) {      
                //logger("DEBUG","main loop");
                
	    	retRead = read_command(answer, sizeof(answer));
            	if (retRead == EOF) {
		    logger("INFO", "Got disconnected");
		    
                    i = 0;
		    k = findItem(EVT_DISCONNECT, &i);
    	    	    if (k && i == 1) {
    	    	    	logger("INFO", "Exec cmd on disconnect");
    	    	    	ret = handleCmdByKey(k);
			// No additional handling since we already disconnected
    	    	    } 

		    if (!dmn && !serverMode) {
                        logger("INFO", "Exit because of disconnect");
                        freeTimers();
		    	freeCfg();
		    	exit(0);
		    }
			
		    if (serverMode == 0) {  // In server mode we will reuse already opened socket
		    	logger("INFO", "Closing the port");
                        closePort(1);
                    }	
		    ret = EXIT_NOK;
		    
		    break;
            	} else if (retRead > 0) {
     	    	    sprintf(tmp, "Command read >%s<", answer);
    	    	    logger("DEBUG",tmp);
	    	
                    ret = handle_command(answer);
		    
		    if (ret == EXIT_OK) {
		    	continue;
		    } else if (ret == EXIT_ABORT) {
                        freeTimers();
		    	freeCfg();
		    	exit(0);
		    } else if (ret == EXIT_DISCON) {
		    	break;
		    }
		}
                
    	    	//Is any alarm fired (timeout about 15 seconds) ?
		if (alarmTimer == 300) {
		    manageAlarms(ALARM_CHECK);
		    alarmTimer = 0;
		} else {
		    alarmTimer++;
		}

    	    	// Is call still active (timeout about 1 seconds) ?
    	    	if (callOn) {
		    if (callTimer >= 20) {
		    	logger("DEBUG", "Get call list again");
			
		    	// Will set global callerId	[MAXLEN];
	    	    	ret = getClip(callerId);
	    	    	if (ret == EXIT_EXACT) {
	    	    	    sprintf(tmp, "Call still active (%s)", callerId);
	    	    	    logger("DEBUG",tmp);
	    	    	} else {
	    	    	    logger("DEBUG","Call was finished");
	    	    	    callOn = 0;
			    strcpy(callerId, "NO CALLER ID");

		    	    int i = 0;
			    type_key *k = findItem(EVT_ENDCALL, &i);
   	    	    	    if (k != NULL && i == 1) {
		    	    	logger("INFO","EVT_ENDCALL found");
  	    	    	    	ret = handleCmdByKey(k);
			    	if (ret == EXIT_ABORT) {
                                    freeTimers();				    
                                    freeCfg();
		    	    	    exit(0);
		    	    	} else if (ret == EXIT_DISCON) {
		    	    	    break;
		    	    	}
	    	    	    }
			}

		    	callTimer = 0;
		    } else {
		    	callTimer++;
		    }
		}
		
		// Should we auto-repeat command (timeout about 1/10 second) ?
		if (repeatNow) {
		    if (repeatTimer == 2) {
		    	logger("DEBUG","Auto repeat command ...");
	    	    	ret = handleCmdByKey(repeatCmdPtr);
	    	    	
			if (ret == EXIT_ABORT) {
                            freeTimers();
			    freeCfg();
		    	    exit(0);
		    	} else if (ret == EXIT_DISCON) {
		    	    break;
		    	}
		    	repeatTimer = 0;
		    } else {
		    	repeatTimer++;
		    }
		}

                // Verify commands executed by timer (timeout about 1 sec)
		if (commandTimer == 20) {
		    //logger("DEBUG","Verify timer commands ...");
	    	    ret = verifyTimers(1) ;
	    	    
		    if (ret == EXIT_ABORT) {
                        freeTimers();
		        freeCfg();
		        exit(0);
		    } else if (ret == EXIT_DISCON) {
		        break;
		    }
		    commandTimer = 0;
		} else {
		    commandTimer++;
		}
		
		// Main loop timer (1/20 of second)
            	usleep(50000);
            }  
    	}
	
	// Can't open port or it closed again
	if (dmn || serverMode > 0 || ret == EXIT_DISCON) {
	    int timeout;
	    
	    if (ret == EXIT_DISCON) {
	    	timeout = TMPDISCONN_TIME;
		ret = EXIT_OK;
	    } else {
	    	timeout = rs;
	    }
	
	    sprintf(tmp, "Wait %d seconds to connect ...", timeout);
	    logger("INFO",tmp);
	    printf("INFO: Wait %d seconds to connect ...\n", timeout);
	    
	    sleep(timeout);
	    
	} else {    // Proceed to exit
	    break;
	}
    }
    
    // Finish all
    logger("INFO", "Stop process ...");
    closePort(1);
    return;
}


int execInFork(char *cmd)
{
    int ret;
    int cpid = fork();
    if( cpid < 0) {
    	logger("ERROR", "Can not fork!");
    	return -1;
    }

    if(cpid) {	    // father
    	waitpid(cpid,&ret,0);
    	if(ret!=-1 && WIFEXITED(ret)) {
	    ret = WEXITSTATUS(ret);
    	}
    } else {	    // child
    	logger("INFO", "Close port in child");
    	closePort(0);
	 
    	ret = system(cmd);
    	if(ret!=-1 && WIFEXITED(ret)) {
    	    ret = WEXITSTATUS(ret);
	}
    	exit(ret);
    }

    return ret;
}


static int execSimpleCmd(char *command, int toFile)
{
    char cmd [1024];
    
    if (command == NULL) {
        logger("ERROR","execSimpleCmd(): null input");
        return EXIT_NOK;
    }
    sprintf(tmp, "execSimpleCmd >%s<", command);
    logger("INFO",tmp);
    
    if (*command == '\0') { // Empty command. Skip it.
    	return EXIT_OK;
    }
    
    if (remoteOn == 1) {
    	char *papameterized = strstr(command, PARAM_CALLID);
	
	if (papameterized) {
    	    *papameterized = '\0';
	    
	    // First part
    	    strcpy(cmd, command);
	    
	    // Add params
	    //strcat(cmd, " ");
	    strcat(cmd, callerId);
	    papameterized += strlen(PARAM_CALLID);
	    
	    // Add second part
	    //strcat(cmd, " ");
	    strcat(cmd, papameterized);
	    sprintf(tmp, "final cmd CALLID >%s<", cmd);
    	    logger("DEBUG",tmp);
	    
	} else if ((papameterized = strstr(command, PARAM_TIME))) {
    	    *papameterized = '\0';
	    
	    // First part
            strcpy(cmd, command);
	    
	    // Add time
	    //strcat(cmd, " ");
	    time_t logtime;
	    time(&logtime);
	
	    struct tm curtime;
	    localtime_r(&logtime, &curtime);
	
	    char buf[64];
	    sprintf(buf,"%d:%02d", curtime.tm_hour, curtime.tm_min);
	
	    papameterized += strlen(PARAM_TIME);
	    strcat(cmd, buf);
	    
	    // Add second part
	    //strcat(cmd, " ");
	    strcat(cmd, papameterized);
	    sprintf(tmp, "final cmd TIME >%s<", cmd);
    	    logger("DEBUG",tmp);
        } else if ((papameterized = strstr(command, PARAM_MODE))) {
	    *papameterized = '\0';
	    
	    // First part
    	    strcpy(cmd, command);
	    
	    // Add params
	    strcat(cmd, getMode());
	    papameterized += strlen(PARAM_MODE);
	    
	    // Add second part
	    strcat(cmd, papameterized);
	    sprintf(tmp, "final cmd MODE >%s<", cmd);
            logger("DEBUG",tmp);
		    
        } else if ((papameterized = strstr(command, PARAM_PARAM))) {
            // This is dirty trick to make Bemused commands like VOLM<volume> works
            
            if (paramValue == -1) {
                logger("ERROR", "Parameter value was not set!");
                return EXIT_OK;
            }
            
	    *papameterized = '\0';
	    
	    // First part
 	    // Add params
	    sprintf(cmd, "%s %d", command, paramValue);
            paramValue = -1; // Reset value after use
	    papameterized += strlen(PARAM_PARAM);
	    
	    // Add second part
	    strcat(cmd, papameterized);
	    sprintf(tmp, "final cmd PARAM >%s<", cmd);
            logger("DEBUG",tmp);
            
        } else {    // Just copy
	    strcpy(cmd, command);
	}

	if (toFile) {       // We needs to wait results, so do not run in background
	    strcat(cmd, " > ");
	    strcat(cmd, TOFILE);
	} else {
	    if (! index(cmd, '&')) {
	        strcat(cmd, "&");
	    }
	}
    	return (/*system(cmd)*/execInFork(cmd) == -1 ? EXIT_NOK : EXIT_OK);
    } else {
    	logger("DEBUG", "Remote if OFF. Skip event");
    }
    
    return EXIT_OK;
}

static int processSendCmd(char *origCmd)
{
    int ret = EXIT_OK;
    
    char all[MAXMAXLEN];
    strncpy(all, origCmd, MAXMAXLEN-1);
    
    char *tag  = strtok(all, ",");
    if (tag == NULL) {
        logger("DEBUG", "processSendCmd(): incorrect input (1)");                
        return EXIT_NOK;
    }
    if (strlen(origCmd) <= strlen(tag) + 1) {   // Set(tag,) or Set(tag) without any value
        logger("DEBUG", "processSendCmd(): incorrect input (2)");                
        return EXIT_NOK;
    }
    char *ptr   = tag + strlen(tag) - 1;
    // strip spaces from tail
    while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
    	*ptr = '\0';
        ptr--;
    }
    ptr = tag + strlen(tag) + 1;
    // strip spaces from head of value
    while( *ptr != '\0' && (*ptr == '\t' || *ptr == ' ')) {
        ptr++;
    }
    
    sprintf(tmp, "processSendCmd() is >%s< >%s<",tag,ptr);
    logger("DEBUG", tmp);
    
    if (strcmp(tag, SEND_STRING) == 0) {
        ret = writeSocketConn(ptr);
        if (ret != EXIT_OK) {
	    logger("DEBUG","Fails in writeSocketConn()");
        }
    } else if (strcmp(tag, SEND_BYTES) == 0) {
        ret = writeBytesSocketConn(ptr);
        if (ret != EXIT_OK) {
	    logger("DEBUG","Fails in writeBytesSocketConn()");
        }    
    } else {
        logger("DEBUG", "processSendCmd(): incorrect input (3)");                
        return EXIT_NOK;
    }
    return ret;
}

int execCmdAndGetResults(int type, char *commandIn, timerCmd* timer) 
{
    logger("DEBUG","execCmdAndGetResults");

    if (serverMode != SERVER_NO && remoteOn == 1 && commandIn != NULL) {
        
        // Parse type of set
        char commandOrig[MAXMAXLEN];
        strncpy(commandOrig, commandIn, MAXMAXLEN-1);
        
        // commandIn = tag,command
        char* tag = strtok(commandOrig,",");
        
        sprintf(tmp, "execCmdAndGetResults() is >%s< >%s<",tag,commandIn);
        logger("DEBUG", tmp);

        if (tag == NULL) {
            logger("DEBUG", "execCmdAndGetResults(): incorrect input (1)");                
            return EXIT_NOK;
        }
        if (strlen(commandIn) <= strlen(tag) + 1) {   // Set(tag,) or Set(tag) without any value
            logger("DEBUG", "execCmdAndGetResults(): incorrect input (2)");                
            return EXIT_NOK;
        }
        if (type != ID_SET && type != ID_SEND ) {
            logger("DEBUG", "execCmdAndGetResults(): incorrect input (3)");                
            return EXIT_NOK;
        }
        char *command2exec = tag + strlen(tag) + 1;
        
        while(*tag == ' ' || *tag == '\t') {
            tag++;
        }
        
        int ret = execSimpleCmd(command2exec,1);
        if (ret != EXIT_OK) {
 	    sprintf(tmp, "Return code from execSimpleCmd() is %d",ret);
 	    logger("DEBUG", tmp);
	    return ret;
        }
        char *res = readResultsFile();
	char *res2;
        int  braces = 1;
        
        if (type == ID_SET) {
            res2  = calloc(strlen(tag) + strlen(res) + strlen(CMD_SET) + 5, sizeof(char));
	    strcpy(res2,CMD_SET);
        } else {                        // ID_SEND - Just send string
            res2  = calloc(strlen(res)+ 1, sizeof(char));
            braces = 0;
        }
	
        if (braces) {
            strcat(res2,"(");
            strcat(res2,tag);
            strcat(res2,",");
        }
	strcat(res2,res);
        if (braces) {
	    strcat(res2,");");
        }
	res[0] = '\0';
	
        int needToWrite = 1;
        if (timer != NULL && timer->lastValue != NULL && strcmp(timer->lastValue,res2) == 0) {
            logger("DEBUG","No needs to send the same data.");
            needToWrite = 0;            // no needs to send exactly the same string
        }
        if (needToWrite) {
            if (type == ID_SEND && strcmp(tag,SEND_BYTES) == 0) {
	        ret = writeBytesSocketConn(res2);
            } else {
	        ret = writeSocketConn(res2);
            }
            if (ret != EXIT_OK) {
	        logger("DEBUG","Fails in write(Bytes)SocketConn()");
            }
            // Store last results
            if (timer != NULL) {
                if (timer->lastValue == NULL) {
                    timer->lastValue = res2;
                } else {
                    free(timer->lastValue);
                    timer->lastValue = res2;
                }
            } else {
                free(res2);res2 = NULL;
            }
        }
        if (timer == NULL) {
            free(res2);res2 = NULL;
        }
    }
       
    return EXIT_OK;
}

static int execCmd(int cmdType, char *commandIn)
{
    int ret;
    char commandCopy[MAXMAXLEN];
    char *commandOrig = NULL;

    if (commandIn == NULL) {
        sprintf(tmp, "execCmd >no params< >%d<", remoteOn);
    } else {
        sprintf(tmp, "execCmd >%s< >%d<", commandIn, remoteOn);
        strncpy(commandCopy, commandIn, MAXMAXLEN-1);
        commandOrig = commandCopy;
    }
    logger("INFO",tmp);
    
    
    switch (cmdType) {
    
        case ID_EXIT:               
    	    if (remoteOn == 1) {	    // User asks to exit	
    	        logger("INFO", "Exit key detected");
    	        finishAll();			    

    	        return EXIT_ABORT;
            }
            break;
            
        case ID_TMPDISCONN:
            if (remoteOn == 1) {            // User asks to temporary disconnect
    	        logger("INFO", "Temporary disconnect key detected");
	        finishAll();
                
    	        return EXIT_DISCON; // But what with the rest of the command line ? (now just skip)
            }
            break;
            
        case ID_REMOTEON:
            logger("INFO", "Switch remote on");
    	    remoteOn = 1;
            break;
            
        case ID_REMOTEOFF:
            logger("INFO", "Switch remote off");
    	    remoteOn = 0;
            break;
            
        case ID_EXEC:
            logger("INFO", "Command: Exec");
            if (remoteOn == 1) {
                execSimpleCmd(commandOrig, 0);
            }
            break;
            
        case ID_SETMODE:
            logger("INFO", "Command: SetMode");
            if (remoteOn == 1) {
                setMode(commandOrig);
            }
            break;

        case ID_SENDCKPD:
            logger("INFO", "Command: SendCKPD");
            if (serverMode == SERVER_NO && remoteOn == 1) {
 
                ret = sendSeq(commandOrig);
                if (ret != EXIT_OK) {
	            logger("DEBUG","Fails in sendSeq()");
                }
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	            return ret;
                }
            }    
            break;

        case ID_SEND:
            logger("INFO", "Command: Send");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                ret = processSendCmd(commandOrig);
                if (ret != EXIT_OK) {
	            logger("DEBUG","Fails in processSendCmd()");
                }
            }    
            break;
            
            
        case ID_SET:          
            logger("INFO", "Command: Set");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                strcpy(tmptext, CMD_SET);
                strcat(tmptext, "(");
                strcat(tmptext, commandOrig);
                strcat(tmptext, ");");
               
                ret = writeSocketConn(tmptext);
                if (ret != EXIT_OK) {
	            logger("DEBUG","Fails in writeSocketConn() (Set)");
                }
            }    
            break;
            
        case ID_EXECSET:   
            logger("INFO", "Command: ExecAndSet");
            
            execCmdAndGetResults(ID_SET, commandOrig, NULL);
            break;
            
        case ID_EXECSEND:  
            logger("INFO", "Command: ExecAndSend");
            
            execCmdAndGetResults(ID_SEND, commandOrig, NULL);
            break;
            
        case ID_TIMERCMD:   
            logger("INFO", "Command: Timer");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                return processTimerCmd(commandOrig);
            }     
            break;

        case ID_CANCELTIMER:   
            logger("INFO", "Command: CancelTimer");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                cancelTimer(commandOrig);
            }  
            break;
            
        default:  
            logger("INFO", "Unknown command");                 
            break;
   }
   
   return EXIT_OK;
}

//
// Functions related to timer commands
//
// Timer(id,timeout,maxtimes,tag,command);
//      where tag could be: icons, title, status
//
int processTimerCmd(char *origCmd)
{
    char all[MAXMAXLEN];
    strcpy(all, origCmd);
    
    char *id  = strtok(all, ",");
    if (id == NULL) {
        logger("DEBUG", "processTimerCmd(): incorrect input - timer data absent");                
        return EXIT_NOK;
    }

    char *ptr = id + strlen(id) - 1;
    // strip spaces from tail
    while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
    	*ptr = '\0';
        ptr--;
    }
    
    char* t1  = strtok(NULL, ",");
    char* t2  = strtok(NULL, ",");
    char* cmd = strtok(NULL, "\0");  
    if (t1 == NULL || t2 == NULL || cmd == NULL) {
        logger("DEBUG", "processTimerCmd(): incorrect input - command is not full");                
        return EXIT_NOK;
    }
    while (*cmd == ' ' || *cmd == '\t') {   // skip spaces
        cmd++;
    }
    sprintf(tmp, "processTimerCmd() cmd = >%s<", cmd);
    logger("DEBUG", tmp);
    
    // Do not allow timer commands for Send(string/bytes...)
    if (strncmp(cmd,SET_ICONS,  strlen(SET_ICONS))  != 0 &&
        strncmp(cmd,SET_TITLE,  strlen(SET_TITLE))  != 0 &&
        strncmp(cmd,SET_STATUS, strlen(SET_STATUS)) != 0) {
        logger("DEBUG", "processTimerCmd(): incorrect tag in command");                
        return EXIT_NOK;
    }
    
    int tmout = atoi(t1);
    int times = atoi(t2);

    if (createTimer(id, ID_SET, tmout, times, cmd) == EXIT_NOK) {
        logger("DEBUG", "processTimerCmd(): timer already exists or input was incorrect");                
        return EXIT_NOK;
    }
    
    timerCmd* tt = NULL;
    execCmdAndGetResults(ID_SET, cmd, findTimer(id, &tt));

    return EXIT_OK;
}

timerCmd* findTimer(char *id, timerCmd **prev) 
{
    timerCmd* ptr = timers;
    
    while (ptr != NULL && strncmp(ptr->id,id,MTEXTLEN) != 0) {
        *prev = ptr;
        //sprintf(tmp, "findTimer(): set prev to >%s<", (*prev)->id);
        //logger("DEBUG",tmp);
        ptr  = ptr->next;
    }

    return ptr;
}

int createTimer(char* id, int type, int tmout, int times, char *cmd)
{
    if (id == NULL || cmd == NULL) {
        logger("DEBUG", "createTimer(): null input");                
        return EXIT_NOK;
    }
    timerCmd* ptr = NULL;
    if (findTimer(id, (timerCmd**) &ptr) != NULL) {
        logger("DEBUG", "createTimer(): timer already exists");                
        return EXIT_NOK;
    } 
    timerCmd *tm = calloc(1,sizeof(timerCmd));
    
    strncpy(tm->id, id, MTEXTLEN);
    tm->type      = type;
    tm->ticks     = 0;
    tm->times     = 1;
    tm->tmout     = tmout;
    tm->maxTimes  = times;
    tm->lastValue = NULL;
    
    tm->cmd = (char *) calloc(1, strlen(cmd) + 1);
    strcpy(tm->cmd, cmd);
    
    // Insert in head
    if (timers == NULL) { // Insert first
        tm->next = NULL;
    } else {
        tm->next = timers;
    }
    timers = tm;
    
    return EXIT_OK;
}

int cancelTimer(char *id) 
{
    sprintf(tmp, "cancelTimer() >%s<", id);
    logger("DEBUG", tmp);
    
    timerCmd *prev = NULL;
    timerCmd *tm = findTimer(id, (timerCmd **) &prev);
    if (tm == NULL) {
        logger("DEBUG", "Cant find timer to cancel!");
        return EXIT_NOK;
    }    
    if (tm == timers) { //remove in head
        timers = tm->next;
    } else {
        if (prev != NULL) {
            prev->next = tm->next;
        } else {
            logger("ERROR", "Previous item absent for non-first element");
            return EXIT_NOK;
        }
    }        

    free(tm->cmd); tm->cmd = NULL;
    free(tm);
    
    return EXIT_OK;
}

void freeTimers()
{
    logger("DEBUG", "freeTimers()");
    timerCmd *tm;
    
    while (timers != NULL) {
        tm     = timers;
        timers = timers->next;
        
        free(tm->cmd); tm->cmd = NULL;
        free(tm);
    }
    return;
}

int verifyTimers(int ticksInSec) 
{
    timerCmd *tm   = timers;
    //timerCmd *prev = NULL;
    
    int numTimers = 0;
    while (tm != NULL) {
        
        if (tm->ticks >= (tm->tmout * ticksInSec)) {
            sprintf(tmp, "verifyTimers(): it is time to execute >%s<", tm->cmd);
            logger("DEBUG",tmp);

            execCmdAndGetResults(tm->type, tm->cmd, tm);
            
            tm->ticks = 0;
            
            if (tm->maxTimes > 0) {
                tm->times++;
                if (tm->times >= tm->maxTimes) {    // Cancel this timer 
                    sprintf(tmp, "verifyTimers(): timer %s self-canceled", tm->id);
                    logger("DEBUG", tmp);
                    
                    // Cancel this timer
                    timerCmd *tmc = tm;
                    
                    // Relink timers list
                    //if (tm == timers) {  // First one
                    //    timers = timers->next;
                    //} else {
                    //    prev->next = tm->next;
                    //}
                    
                    tm = tm->next;

                    cancelTimer(tmc->id);
                    
                    continue;
                }
            }
        } else {
            tm->ticks++;
        }
        //prev = tm;
        tm   = tm->next;
        numTimers++;
    }
    return numTimers;
}

//
// Bemused server limited emulation
// we could get more than one command at once
//
      
static int isBemusedCommand(char *cmdIn)
{  
    sprintf(tmp, "isBemusedCommand: >%s<",cmdIn);
    logger("DEBUG",tmp);
    
    int isBemused = 0;
    
    char *cmd = cmdIn;
    char *last = cmdIn + strlen(cmdIn);
    
    while (cmd < last) {
    	int handle = 0;
	int shift  = 4;
	char oneCmd[5];
	
    	if (strncmp(cmd, "CHCK", 4) == 0 ||
	    strncmp(cmd, "DINF", 4) == 0 ||
    	    strncmp(cmd, "EXIT", 4) == 0 ||
    	    strncmp(cmd, "FADE", 4) == 0 ||
    	    strncmp(cmd, "FFWD", 4) == 0 ||
	    strncmp(cmd, "GVOL", 4) == 0 ||
	    strncmp(cmd, "INF2", 4) == 0 ||
	    strncmp(cmd, "INFO", 4) == 0 ||
	    strncmp(cmd, "LIST", 4) == 0 ||
    	    strncmp(cmd, "NEXT", 4) == 0 ||
    	    strncmp(cmd, "PAUS", 4) == 0 ||
	    strncmp(cmd, "PLEN", 4) == 0 ||
	    strncmp(cmd, "PLST", 4) == 0 ||
    	    strncmp(cmd, "PREV", 4) == 0 ||
	    strncmp(cmd, "RMAL", 4) == 0 ||
    	    strncmp(cmd, "RWND", 4) == 0 ||
	    strncmp(cmd, "SHUT", 4) == 0 ||
    	    strncmp(cmd, "STEN", 4) == 0 ||
    	    strncmp(cmd, "STOP", 4) == 0 ||
            strncmp(cmd, "STRT", 4) == 0 ||
	    strncmp(cmd, "VERS", 4) == 0) {
	    handle = 1;
    	}

    	if (strncmp(cmd, "REPT", 4) == 0 ||
            strncmp(cmd, "SHFL", 4) == 0) {
            unsigned char bc = ((unsigned char) *(cmd + 4));
            paramValue = (int)bc;

            handle = 1;
	    shift  = 5;
    	}
    	if (strncmp(cmd, "VOLM", 4) == 0) {
	    unsigned char bc;
	    if (strlen(cmd) == 4) {  // read only VOLM without value to set
	    	logger("DEBUG","read one more byte");
		char ret;
	    	//if ((ret = readSocketConn((char*)buf, 2)) != 1) {
	    	if ((ret = bt_readchar(500000)) < 0) {
            	    sprintf(tmp, "isBemusedCommand: read < 0");
            	    logger("DEBUG",tmp);
		    paramValue = -1;
		    return 0;
		}
            	sprintf(tmp, "isBemusedCommand: read >%c<", ret);
            	logger("DEBUG",tmp);
		bc = ((unsigned char) ret);
	    } else {
	    	bc = ((unsigned char) *(cmd + 4));
	    }
 
    	    // VOLM 0-255 -> 0%-100%
            paramValue = ((int)bc * 100)/255;

        
            sprintf(tmp, "isBemusedCommand: parameter >%d<", paramValue);
            logger("DEBUG",tmp);
	    
	    handle = 1;
	    shift  = 5;
    	}
    	if (strncmp(cmd, "SLCT", 4) == 0) {
            handle = 1;
	    shift  = 6;
    	}
    	if (strncmp(cmd, "SEEK", 4) == 0)  {	
	
	    paramValue = (((unsigned char)*(cmd + 4)) << 24) + 
	                 (((unsigned char)*(cmd + 5)) << 16) + 
		         (((unsigned char)*(cmd + 6)) << 8)  + 
		           (unsigned char)*(cmd + 7);
            handle = 1;
	    shift  = 8;
    	}
    	if (strncmp(cmd, "DLST", 4) == 0 || 
            strncmp(cmd, "DOWN", 4) == 0 || 
            strncmp(cmd, "FINF", 4) == 0 ||    
    	    strncmp(cmd, "LADD", 4) == 0 ||
    	    strncmp(cmd, "PLAY", 4) == 0) {             
	    
	    unsigned char bc = *((unsigned char*) (cmd + 4));
	    
	    shift  = 5 + bc;
            handle = 1;
    	}
    	if (handle) {
	    isBemused = 1;
	    strncpy(oneCmd,cmd,4);
	    oneCmd[4] = '\0';
           
	    sprintf(tmp, "isBemusedCommand: one commmand >%s<", oneCmd);
            logger("DEBUG",tmp);

    	    
    	    handle_key_press(oneCmd);
    	} else {
	    break;
	}
	
	cmd = cmd + shift;
    }
    return isBemused;
}
